import jax.numpy as jnp
from numpy.testing import assert_array_equal

from reservoirpy.datasets import mackey_glass
from reservoirpy.jax.nodes import LIF


def test_lif():
    n_timesteps = 140
    neurons = 100

    lif = LIF(
        units=neurons,
        inhibitory=0.0,
        sr=1.0,
        lr=0.2,
        input_scaling=1.0,
        threshold=1.0,
        rc_connectivity=1.0,
    )

    x = jnp.ones((n_timesteps, 1))
    y = lif.run(x)

    assert y.shape == (n_timesteps, neurons)
    assert_array_equal(jnp.sort(jnp.unique(y)), jnp.array([0.0, 1.0]))



# # Licence: MIT License
# # Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>

# import jax.numpy as jnp
# import numpy
# import pytest

# from reservoirpy.jax.mat_gen import ring
# from reservoirpy.jax.nodes import LocalPlasticityReservoir


# def test_lsp_init():
#     x = jnp.ones((10, 5))
#     res = LocalPlasticityReservoir(100, input_dim=5)

#     res.initialize(x)

#     assert res.W.shape == (100, 100)
#     assert res.Win.shape == (100, 5)

#     res = LocalPlasticityReservoir(100)

#     out = res.run(x)

#     assert out.shape == (10, 100)
#     assert res.W.shape == (100, 100)
#     assert res.Win.shape == (100, 5)

#     with pytest.raises(ValueError):
#         _ = LocalPlasticityReservoir(local_rule="oja")


# def test_lsp_rules():
#     x = jnp.ones((10, 5))

#     res = LocalPlasticityReservoir(20, local_rule="oja")
#     _ = res.fit(x)
#     res = LocalPlasticityReservoir(100, local_rule="anti-oja")
#     _ = res.fit(x)
#     res = LocalPlasticityReservoir(100, local_rule="hebbian")
#     _ = res.fit(x)
#     res = LocalPlasticityReservoir(100, local_rule="anti-hebbian")
#     _ = res.fit(x)
#     res = LocalPlasticityReservoir(100, local_rule="bcm")
#     _ = res.fit(x)

#     with pytest.raises(ValueError):
#         res = LocalPlasticityReservoir(100, local_rule="anti-bcm")

#     with pytest.raises(ValueError):
#         res = LocalPlasticityReservoir(100, local_rule="anti_oja")


# def test_local_synaptic_plasticity():
#     rng = numpy.random.default_rng(seed=1)
#     x = rng.normal(size=(100, 5))
#     X = [x[:10], x[:20]]

#     res = LocalPlasticityReservoir(100, local_rule="hebbian", epochs=2, seed=0)

#     res.fit(x)
#     res.fit(X)

#     assert res.W.shape == (100, 100)

#     res = LocalPlasticityReservoir(100, local_rule="oja", epochs=10, eta=1e-3, synapse_normalization=True, seed=10)
#     res.initialize(x)

#     initial_Wvals = res.W.copy()

#     res.fit(x)
#     res.fit(X)

#     assert not jnp.allclose(initial_Wvals, res.W)


# def test_lsp_matrices():
#     rng = numpy.random.default_rng(seed=2504)
#     x = rng.normal(size=(100, 5))

#     W = jnp.array(ring(10, 10))

#     lspres_ring = LocalPlasticityReservoir(W=W, seed=2504)
#     lspres_rand = LocalPlasticityReservoir(units=10, seed=2504)

#     lspres_ring.fit(x)
#     lspres_rand.fit(x)

#     assert not jnp.allclose(lspres_ring.W, lspres_rand.W)

#     # test dense matrix
#     W = jnp.array(rng.normal(size=(10, 10)))
#     res = LocalPlasticityReservoir(W=W)
#     res.fit(x)



# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>

from typing import Optional, Union

import jax
import numpy as np

from ..utils.random import rand_generator


def rng_key(seed: Optional[Union[int, np.random.Generator]] = None) -> "jax.random.PRNGKey":
    if seed is None or isinstance(seed, np.random.Generator):
        return rand_generator(seed).integers(1 << 64 - 1)
    return jax.random.key(seed)



# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>

import jax.numpy as jnp
import numpy
import pytest
from joblib import Parallel, delayed
from numpy.testing import assert_array_almost_equal, assert_array_equal

from reservoirpy.jax.nodes import Ridge


def test_ridge_init():
    x: jax.Array = jnp.ones((100,))
    X: jax.Array = jnp.ones((120, 100))
    Y: jax.Array = jnp.ones((120, 10))

    node = Ridge(ridge=1e-8)
    node.initialize(X, Y)
    assert node.output_dim == 10
    assert node.input_dim == 100
    assert node.ridge == 1e-8

    node = Ridge(ridge=1e-8, output_dim=10)
    node.initialize(X)
    assert node.output_dim == 10
    assert node.input_dim == 100
    assert node.ridge == 1e-8

    with pytest.raises((ValueError, TypeError)):
        node = Ridge(ridge=1e-8, input_dim=100, output_dim=10)
        node(x)

    node = Ridge(ridge=1e-8, Wout=jnp.ones((100, 10)), bias=jnp.ones((10,)))
    node.run(X)
    assert node.output_dim == 10
    assert node.input_dim == 100
    assert node.ridge == 1e-8


def test_ridge_worker():
    X: jax.Array = 2 * jnp.ones((120, 100))
    Y: jax.Array = 2 * jnp.ones((120, 10))

    node = Ridge(1e-6)

    XXT, YXT, x_sum, y_sum, sample_size = node.worker(X, Y)
    assert XXT.shape == (100, 100)
    assert YXT.shape == (100, 10)
    assert_array_equal(x_sum, 2 * 120 * jnp.ones((100,)))
    assert_array_equal(y_sum, 2 * 120 * jnp.ones((10,)))
    assert_array_equal(sample_size, 120)


def test_ridge_fit():
    rng = numpy.random.default_rng(seed=0)
    X: jax.Array = rng.uniform(size=(12, 100))
    Y: jax.Array = X @ rng.uniform(size=(100, 10)) + rng.uniform(size=10)

    node = Ridge(1e-6)
    node.fit(X, Y)
    assert node.input_dim == 100
    assert node.output_dim == 10
    assert node.Wout.shape == (100, 10)
    assert node.bias.shape == (10,)
    assert jnp.any(node.bias != jnp.zeros((10,)))

    node = Ridge(1e-6, fit_bias=False)
    node.fit(X, Y)
    assert node.input_dim == 100
    assert node.output_dim == 10
    assert node.Wout.shape == (100, 10)
    assert_array_equal(node.bias, jnp.zeros((10,)))


def test_ridge_fit_multiseries():
    rng = numpy.random.default_rng(seed=0)
    X: jax.Array = rng.uniform(size=(15, 12, 100))
    Y: jax.Array = X @ rng.uniform(size=(100, 10)) + rng.uniform(size=10)

    node = Ridge(1e-6)
    node.fit(X, Y)
    assert node.input_dim == 100
    assert node.output_dim == 10
    assert node.Wout.shape == (100, 10)
    assert node.bias.shape == (10,)
    assert not jnp.all(node.Wout == jnp.zeros((100, 10)))
    assert not jnp.all(node.bias == jnp.zeros((10,)))
    node2 = Ridge(1e-6)
    node2.fit(X.reshape(15 * 12, 100), Y.reshape(15 * 12, 10))
    assert_array_almost_equal(node.Wout, node2.Wout, decimal=2)
    assert_array_almost_equal(node.bias, node2.bias, decimal=2)

    node = Ridge(1e-6, fit_bias=False)
    node.fit(X, Y)
    assert node.input_dim == 100
    assert node.output_dim == 10
    assert node.Wout.shape == (100, 10)
    assert_array_equal(node.bias, jnp.zeros((10,)))


def test_ridge_fit_parallel():
    rng = numpy.random.default_rng(seed=0)
    X: jax.Array = rng.uniform(size=(15, 12, 100))
    Y: jax.Array = X @ rng.uniform(size=(100, 10)) + rng.uniform(size=10)

    node = Ridge(1e-6)
    node.fit(X, Y, workers=-1)
    assert node.input_dim == 100
    assert node.output_dim == 10
    assert node.Wout.shape == (100, 10)
    assert node.bias.shape == (10,)
    assert jnp.any(node.bias != jnp.zeros((10,)))

    node = Ridge(1e-6, fit_bias=False)
    node.fit(X, Y)
    assert node.input_dim == 100
    assert node.output_dim == 10
    assert node.Wout.shape == (100, 10)
    assert_array_equal(node.bias, jnp.zeros((10,)))



# This is much more simple than reservoirpy.activationsfunc
# since those activation functions are defined in jax.nn

from typing import Callable, Union

from jax.nn import identity, relu, sigmoid, softmax, softplus, tanh


def get_function(name: Union[Callable, str]) -> Callable:
    """Return a Jax activation function from name.

    Parameters
    ----------
    name : str, Callable
        Name of the activation function.
        Can be one of {'softmax', 'softplus',
        'sigmoid', 'tanh', 'identity', 'relu'} or
        their respective short names {'smax', 'sp',
        'sig', 'id', 're'}. If `name` is a Callable,
        simply returns `name`.

    Returns
    -------
    callable
        An activation function.
    """
    if isinstance(name, Callable):
        return name

    index = {
        "softmax": softmax,
        "softplus": softplus,
        "sigmoid": sigmoid,
        "tanh": tanh,
        "identity": identity,
        "relu": relu,
        "smax": softmax,
        "sp": softplus,
        "sig": sigmoid,
        "id": identity,
        "re": relu,
    }

    if not name in index:
        raise ValueError(f"Function name must be one of {[k for k in index.keys()]}")
    else:
        return index[name]



# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>

import numpy as np
import pytest
from joblib import Parallel, delayed
from numpy.testing import assert_array_almost_equal, assert_array_equal

from reservoirpy.nodes import Ridge


def test_ridge_init():
    x: np.ndarray = np.ones((100,))
    X: np.ndarray = np.ones((120, 100))
    Y: np.ndarray = np.ones((120, 10))

    node = Ridge(ridge=1e-8)
    node.initialize(X, Y)
    assert node.output_dim == 10
    assert node.input_dim == 100
    assert node.ridge == 1e-8

    node = Ridge(ridge=1e-8, output_dim=10)
    node.initialize(X)
    assert node.output_dim == 10
    assert node.input_dim == 100
    assert node.ridge == 1e-8

    with pytest.raises(ValueError):
        node = Ridge(ridge=1e-8, input_dim=100, output_dim=10)
        node(x)

    node = Ridge(ridge=1e-8, Wout=np.ones((100, 10)), bias=np.ones((10,)))
    node.run(X)
    assert node.output_dim == 10
    assert node.input_dim == 100
    assert node.ridge == 1e-8


def test_ridge_worker():
    X: np.ndarray = 2 * np.ones((120, 100))
    Y: np.ndarray = 2 * np.ones((120, 10))

    node = Ridge(1e-6)

    XXT, YXT, x_sum, y_sum, sample_size = node.worker(X, Y)
    assert XXT.shape == (100, 100)
    assert YXT.shape == (100, 10)
    assert_array_equal(x_sum, 2 * 120 * np.ones((100,)))
    assert_array_equal(y_sum, 2 * 120 * np.ones((10,)))
    assert_array_equal(sample_size, 120)


def test_ridge_fit():
    X: np.ndarray = np.ones((120, 100))
    Y: np.ndarray = np.ones((120, 10))

    node = Ridge(1e-6)
    node.fit(X, Y)
    assert node.input_dim == 100
    assert node.output_dim == 10
    assert node.Wout.shape == (100, 10)
    assert node.bias.shape == (10,)
    assert np.any(node.bias != np.zeros((10,)))

    node = Ridge(1e-6, fit_bias=False)
    node.fit(X, Y)
    assert node.input_dim == 100
    assert node.output_dim == 10
    assert node.Wout.shape == (100, 10)
    assert_array_equal(node.bias, np.zeros((10,)))


def test_ridge_fit_multiseries():
    rng = np.random.default_rng(seed=0)
    X: np.ndarray = rng.uniform(size=(15, 12, 100))
    Y: np.ndarray = X @ rng.uniform(size=(100, 10)) + rng.uniform(size=10)

    node = Ridge(1e-6)
    node.fit(X, Y)
    assert node.input_dim == 100
    assert node.output_dim == 10
    assert node.Wout.shape == (100, 10)
    assert node.bias.shape == (10,)
    assert not np.all(node.Wout == np.zeros((100, 10)))
    assert not np.all(node.bias == np.zeros((10,)))
    node2 = Ridge(1e-6)
    node2.fit(X.reshape(15 * 12, 100), Y.reshape(15 * 12, 10))
    assert_array_almost_equal(node.Wout, node2.Wout)
    assert_array_almost_equal(node.bias, node2.bias)

    node = Ridge(1e-6, fit_bias=False)
    node.fit(X, Y)
    assert node.input_dim == 100
    assert node.output_dim == 10
    assert node.Wout.shape == (100, 10)
    assert_array_equal(node.bias, np.zeros((10,)))


def test_ridge_fit_parallel():
    X: np.ndarray = np.ones((15, 12, 100))
    Y: np.ndarray = np.ones((15, 12, 10))

    node = Ridge(1e-6)
    node.fit(X, Y, workers=-1)
    assert node.input_dim == 100
    assert node.output_dim == 10
    assert node.Wout.shape == (100, 10)
    assert node.bias.shape == (10,)
    assert np.any(node.bias != np.zeros((10,)))

    node = Ridge(1e-6, fit_bias=False)
    node.fit(X, Y)
    assert node.input_dim == 100
    assert node.output_dim == 10
    assert node.Wout.shape == (100, 10)
    assert_array_equal(node.bias, np.zeros((10,)))



# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from . import activationsfunc, mat_gen, nodes, type, utils
from .esn import ESN
from .model import Model
from .node import Node

__all__ = [
    "mat_gen",
    "activationsfunc",
    # "observables",  # NOT PLANNED FOR NOW
    # "hyper",  # NOT PLANNED
    # "datasets",  # NOT PLANNED
    "nodes",
    "Node",
    "Model",
    "ESN",
    "type",
    "utils",
]


def __getattr__(attr):
    if attr in ["observables", "hyper", "datasets", "set_seed", "__version__"]:
        raise AttributeError(f"{attr} does not have a Jax implementation. Please use reservoir.{attr} instead.")
    else:
        raise AttributeError(f"module {__name__!r} has no attribute {attr}")



# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>

from functools import partial
from typing import Callable, Iterable, Optional, Union

import jax
import jax.numpy as jnp
from scipy import linalg

from ..node import ParallelNode
from ..type import NodeInput, State, Timeseries, Timestep, Weights, is_array


class Ridge(ParallelNode):
    """A single layer of neurons learning with Tikhonov linear regression.

    Output weights of the layer are computed following:

    .. math::

        \\hat{\\mathbf{W}}_{out} = \\mathbf{YX}^\\top ~ (\\mathbf{XX}^\\top +
        \\lambda\\mathbf{Id})^{-1}

    Outputs :math:`\\mathbf{y}` of the node are the result of:

    .. math::

        \\mathbf{y} = \\mathbf{W}_{out}^\\top \\mathbf{x} + \\mathbf{b}

    where:
        - :math:`\\mathbf{X}` is the accumulation of all inputs during training;
        - :math:`\\mathbf{Y}` is the accumulation of all targets during training;
        - :math:`\\mathbf{b}` is the first row of :math:`\\hat{\\mathbf{W}}_{out}`;
        - :math:`\\mathbf{W}_{out}` is the rest of :math:`\\hat{\\mathbf{W}}_{out}`.

    If ``fit_bias`` is True, then :math:`\\mathbf{b}` is non-zero, and a constant
    term is added to :math:`\\mathbf{X}` to compute it.


    Parameters
    ----------
    ridge: float, default to 0.0
        L2 regularization parameter.
    fit_bias : bool, default to True
        If True, then a bias parameter will be learned along with output weights.
    Wout : callable or array-like of shape (input_dim, units), optional
        Output weights matrix or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    bias : callable or array-like of shape (units,), optional
        Bias weights vector or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    input_dim : int, optional
        Input dimension. Can be inferred at first call.
    output_dim : int, optional
        Number of units in the readout, can be inferred at first call.
    name : str, optional
        Node name.

    Example
    -------

    >>> x = np.random.normal(size=(100, 3))
    >>> noise = np.random.normal(scale=0.1, size=(100, 1))
    >>> y = x @ np.array([[10], [-0.2], [7.]]) + noise + 12.
    >>>
    >>> from reservoirpy.nodes import Ridge
    >>> ridge_regressor = Ridge(ridge=0.001)
    >>>
    >>> ridge_regressor.fit(x, y)
    >>> ridge_regressor.Wout, ridge_regressor.bias
    array([[ 9.992, -0.205,  6.989]]).T, array([[12.011]])
    """

    #: Regularization parameter (:math:`\\lambda`) (0.0 by default).
    ridge: float
    #: If True, learn a bias term (True by default).
    fit_bias: bool
    #: Learned output weights (:math:`\\mathbf{W}_{out}`).
    Wout: Weights
    #: Learned bias (:math:`\\mathbf{b}`).
    bias: Weights

    def __init__(
        self,
        ridge: float = 0.0,
        fit_bias: bool = True,
        Wout: Optional[Union[Weights, Callable]] = None,
        bias: Optional[Union[Weights, Callable]] = None,
        input_dim: Optional[int] = None,
        output_dim: Optional[int] = None,
        name: Optional[str] = None,
    ):
        self.ridge = ridge
        self.fit_bias = fit_bias
        self.Wout = Wout
        self.bias = bias
        self.name = name
        self.state = {}

        # set input_dim/output_dim (if possible)
        self.input_dim = input_dim
        self.output_dim = output_dim
        if is_array(Wout):
            if input_dim is not None and Wout.shape[0] != input_dim:
                raise ValueError(
                    f"Both 'input_dim' and 'Wout' are set but their dimensions doesn't "
                    f"match: {input_dim} != {Wout.shape[0]}."
                )
            self.input_dim = Wout.shape[0]
            if output_dim is not None and Wout.shape[1] != output_dim:
                raise ValueError(
                    f"Both 'output_dim' and 'Wout' are set but their dimensions doesn't "
                    f"match: {output_dim} != {Wout.shape[1]}."
                )
            self.output_dim = Wout.shape[1]
        if is_array(bias):
            if output_dim is not None and bias.shape[0] != output_dim:
                raise ValueError(
                    f"Both 'output_dim' and 'Wout' are set but their dimensions doesn't "
                    f"match: {output_dim} != {bias.shape[0]}."
                )
            self.output_dim = bias.shape[0]

    def initialize(
        self,
        x: Union[NodeInput, Timestep],
        y: Optional[Union[NodeInput, Timestep]] = None,
    ):
        self._set_input_dim(x)
        self._set_output_dim(y)

        # initialize matrices
        if isinstance(self.Wout, Callable):
            self.Wout = self.Wout(self.input_dim, self.output_dim)
        if isinstance(self.bias, Callable):
            self.bias = self.bias(self.output_dim)
        self.state = {"out": jnp.zeros((self.output_dim,))}

        self.initialized = True

    @partial(jax.jit, static_argnums=(0,))
    def _step(self, state: State, x: Timestep) -> State:
        return {"out": x @ self.Wout + self.bias}

    def _run(self, state: State, x: Timeseries) -> tuple[State, Timeseries]:
        out = x @ self.Wout + self.bias
        return {"out": out[-1]}, out

    def worker(self, x: Timeseries, y: Timeseries):
        x_sum = jnp.sum(x, axis=0)
        y_sum = jnp.sum(y, axis=0)
        sample_size = x.shape[0]
        XXT = x.T @ x
        YXT = x.T @ y
        return XXT, YXT, x_sum, y_sum, sample_size

    def master(self, generator: Iterable):
        xxt, yxt, x_sum, y_sum, sample_size = generator
        XXT = jnp.sum(xxt, axis=0)
        YXT = jnp.sum(yxt, axis=0)
        X_sum = jnp.sum(x_sum, axis=0)
        Y_sum = jnp.sum(y_sum, axis=0)
        total_samples = jnp.sum(sample_size)
        ridge_In = self.ridge * jnp.eye(self.input_dim)

        if self.fit_bias:
            X_means = X_sum / total_samples
            Y_means = Y_sum / total_samples
            XXT -= total_samples * jnp.outer(X_means, X_means)
            YXT -= total_samples * jnp.outer(X_means, Y_means)

        Wout = linalg.solve(XXT + ridge_In, YXT, assume_a="sym")

        self.Wout = Wout
        if self.fit_bias:
            self.bias = Y_means - X_means @ Wout
        else:
            self.bias = jnp.zeros((self.output_dim,))



# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from .random import rand_generator

EXCLUDED_PARAMETERS = ["Win", "W", "bias", "Wout"]


def get_non_defaults(instance, constructor=None):
    """
    Return a dictionary of key: values where the key is both an attribute of ``instance`` and a named parameter of its
    constructor (or a parameter of ``constructor`` if specified), but the attribute value is different from the default
    value of the constructor. For the sake of readability, some parameters names are excluded, such as ``W`` or ``bias``.

    This method is used for the ``__str__`` and ``__repr__`` methods.

    Parameters
    ----------
    instance : object
    constructor : class, optional

    Returns
    -------
    dict[str, Any]
    """
    if constructor is None:
        constructor = type(instance)
    import inspect

    signature = inspect.signature(constructor)
    defaults = {k: v.default for k, v in signature.parameters.items() if v.default is not inspect.Parameter.empty}
    return {
        k: getattr(instance, k)
        for k, v in defaults.items()
        if hasattr(instance, k) and k not in EXCLUDED_PARAMETERS and getattr(instance, k) != v
    }



"""
===================================
Models (:class:`reservoirpy.Model`)
===================================

Note
----

See the following guides to:

- **Learn more about how to work with ReservoirPy Nodes**: :ref:`node`
- **Learn more about how to combine nodes within a Model**: :ref:`model`


Models are an extension of the Node API. They allow to combine nodes into
complex computational graphs, to create complicated Reservoir Computing
architecture like *Deep Echo State Networks*.

See :ref:`model` to learn more about how to create and manipulate
a :py:class:`Model`.

.. currentmodule:: reservoirpy.model

.. autoclass:: Model

   .. rubric:: Methods

   .. autosummary::

      ~Model.initialize
      ~Model.step
      ~Model.run
      ~Model.predict
      ~Model.fit
      ~Model.partial_fit


   .. rubric:: Attributes

   .. autosummary::

    ~Model.nodes
    ~Model.edges
    ~Model.inputs
    ~Model.outputs
    ~Model.named_nodes
    ~Model.trainable_nodes
    ~Model.execution_order
    ~Model.parents
    ~Model.children
    ~Model.is_trainable
    ~Model.is_multi_input
    ~Model.is_multi_output
    ~Model.is_parallel
    ~Model.initialized


"""

# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>

from collections import defaultdict
from typing import Mapping, Optional, Sequence, Union

import numpy as np
from joblib import Parallel, delayed

from reservoirpy.node import Node
from reservoirpy.type import FeedbackBuffers, NodeInput, State, Timestep
from reservoirpy.utils.data_validation import check_model_input, check_model_timestep

from .node import Node, OnlineNode, ParallelNode, TrainableNode
from .type import (
    FeedbackBuffers,
    ModelInput,
    ModelTimestep,
    NodeInput,
    State,
    Timeseries,
    Timestep,
    get_data_dimension,
    is_multiseries,
)
from .utils.graphflow import (
    find_indirect_children,
    find_inputs,
    find_outputs,
    find_parents_and_children,
    find_pseudo_inputs,
    topological_sort,
    unique_ordered,
)
from .utils.model_utils import (
    check_input_output_connections,
    check_unnamed_in_out,
    check_unnamed_trainable,
    data_from_buffer,
    join_data,
    map_input,
    map_teacher,
    mapping_iterator,
    unfold_mapping,
)


class Model:
    """Model base class.

    Parameters
    ----------
    nodes : list of Node, optional
        Nodes to include in the Model.
    edges : list of (Node, int, Node), optional
        Edges between Nodes in the graph. An edge between a
        Node A and a Node B with a delay of :math:`d` is created as a tuple ``(A, d, B)``.
    """

    #: List of Nodes contained in the model, in insertion order.
    nodes: list[Node]
    #: List of ``(NodeA, d, NodeB)`` edges, representing a connection from ``NodeA`` to ``NodeB`` with a delay of ``d``.
    edges: list[tuple[Node, int, Node]]
    #: Dictionary of edges -> np.ndarray where arrays contain the values to be sent to the receiving node.
    feedback_buffers: FeedbackBuffers
    #: List of Nodes that expects an input
    inputs: list[Node]
    #: List of Nodes expected to output their values
    outputs: list[Node]
    #: Dictionary that associates a name to a Node with that name in the model.
    named_nodes: dict[str, Node]
    #: List of nodes that can be trained
    trainable_nodes: list[Node]
    #: List of Nodes contained in the model, in the order they should be run.
    execution_order: list[Node]
    #: Dictionary that associates a list of all Nodes connected to the key Node.
    parents: dict[Node, list[Node]]
    #: Dictionary that associates a list of all Nodes to which the key node is connected.
    children: dict[Node, list[Node]]

    #: If True, the model can be trained (with :py:meth:`~.Model.fit`).
    is_trainable: bool
    #: If True, the model has multiple Node inputs
    is_multi_input: bool
    #: If True, the model has multiple outputs and returns a dictionary that associates a node name to a node output.
    is_multi_output: bool
    #: If True, the model can be trained online (with :py:meth:`~.Model.partial_fit`).
    is_online: bool
    #: If True, the model can be trained in parallel (offline).
    is_parallel: bool
    #: If True, the model and its Nodes has been initialized.
    initialized: bool

    def __init__(
        self,
        nodes: Sequence[Node],
        edges: Sequence[tuple[Node, int, Node]],
    ):
        # convert to List[Node], removes duplicates, use dict to preserve order
        self.nodes = unique_ordered(nodes)
        self.edges = unique_ordered(edges)

        self.inputs = find_inputs(self.nodes, self.edges)
        self.outputs = find_outputs(self.nodes, self.edges)
        self.named_nodes = {n.name: n for n in self.nodes if n.name is not None}
        self.trainable_nodes = [n for n in nodes if isinstance(n, TrainableNode)]
        self.is_trainable = len(self.trainable_nodes) > 0
        self.is_multi_input = len(self.inputs) > 1
        self.is_multi_output = len(self.outputs) > 1
        self.is_online = all([isinstance(n, OnlineNode) for n in self.trainable_nodes])
        self.is_parallel = all([isinstance(n, ParallelNode) for n in self.trainable_nodes])
        self.parents, self.children = find_parents_and_children(self.nodes, self.edges)

        # execution order / cycle detection (without teacher forcing)
        direct_edges = [edge for edge in self.edges if edge[1] == 0]
        self.execution_order = topological_sort(self.nodes, direct_edges, inputs=self.inputs)
        self.feedback_buffers = None

        self.initialized = False

    def initialize(
        self,
        x: Union[ModelInput, ModelTimestep],
        y: Optional[Union[ModelInput, ModelTimestep]] = None,
    ):
        """Initializes a :py:class:`Model` instance at runtime, using samples of
        data to infer all :py:class:`Node` dimensions and instantiate the feedback buffers.

        Parameters
        ----------
        x : numpy.ndarray or dict of numpy.ndarray
            A vector of shape `(1, ndim)` corresponding to a timestep of data, or
            a dictionary mapping node names to vector of shapes
            `(1, ndim of corresponding node)`.
        y : numpy.ndarray or dict of numpy.ndarray, optional
            A vector of shape `(1, ndim)` corresponding to a timestep of target
            data, or a dictionary mapping node names to vector of
            shapes `(1, ndim of corresponding node)`.
        """
        check_unnamed_in_out(self)
        check_input_output_connections(self.edges)
        check_unnamed_trainable(self)

        y_ = map_teacher(self, y)

        # Infer node input dimensions from the input they receive

        node_input_dims = {node: 0 for node in self.nodes}

        if isinstance(x, dict):
            for node_name, val in x.items():
                node = self.named_nodes[node_name]
                node_input_dims[node] += get_data_dimension(val)
        else:
            for node in self.inputs:
                node_input_dims[node] += get_data_dimension(x)

        # also use y as forced teachers. Useful for models with feedback
        indirect_children = find_indirect_children(nodes=self.nodes, edges=self.edges)
        for supervised_node, y_teacher in y_.items():
            for child in indirect_children[supervised_node]:
                node_input_dims[child] += get_data_dimension(y_teacher)

        # execution order / cycle detection (with teacher forcing)
        pseudo_inputs = find_pseudo_inputs(self.nodes, self.edges, y_mapping=y_)
        pseudo_edges = [edge for edge in self.edges if edge[0] not in y_]
        self.pseudo_execution_order = topological_sort(self.nodes, pseudo_edges, inputs=pseudo_inputs)
        # Initialize each node in execution_order
        for node in self.pseudo_execution_order:
            node_input_dim = node_input_dims[node]
            if node.initialized:
                if node_input_dim != node.input_dim:
                    raise ValueError(
                        f"{node} expects input of dimension {node.input_dim} "
                        f"but receives input of dimension {node_input_dim}."
                    )
            else:
                if node in y_:
                    node.initialize(x=np.zeros((node_input_dim,)), y=y_[node])
                else:
                    node.initialize(x=np.zeros((node_input_dim,)))
            if node in y_.keys():
                if get_data_dimension(y_[node]) != node.output_dim:
                    raise ValueError(
                        f"{node} expects training data of dimension {node.output_dim} "
                        f"but receives data of dimension {get_data_dimension(y_[node])}."
                    )
            else:
                for child in self.children[node]:
                    node_input_dims[child] += node.output_dim

        self.feedback_buffers = {(p, d, c): np.zeros((d, p.output_dim)) for p, d, c in self.edges if d > 0}

        # TODO: Jax compilation
        self.initialized = True

    def _step(
        self,
        state: tuple[FeedbackBuffers, Mapping[Node, State]],
        x: Mapping[Node, Timestep],
    ) -> tuple[FeedbackBuffers, dict[Node, State]]:
        buffers, node_states = state

        new_state: dict[Node, State] = {}

        for node in self.execution_order:
            inputs = []
            if node in x:
                inputs.append(x[node])
            inputs += [new_state[parent]["out"] for parent in self.parents[node]]
            inputs += [buffer[-1] for (_p, _d, c), buffer in buffers.items() if c == node]
            node_input = np.concatenate(inputs, axis=-1)
            new_state[node] = node._step(node_states[node], node_input)

        new_buffers = {edge: buffer.copy() for edge, buffer in buffers.items()}
        for (p, d, c), buffer in new_buffers.items():
            buffer[-1] = new_state[p]["out"]
            new_buffers[(p, d, c)] = np.roll(buffer, 1, axis=0)

        return new_buffers, new_state

    def step(self, x: Optional[ModelTimestep] = None) -> ModelTimestep:
        """Call the Model function on a single step of data and update
        its state.

        Parameters
        ----------
        x : array of shape (input_dim,), optional
            One single step of input data. If None, an empty array is used
            instead and the Node is assumed to have an input_dim of 0

        Returns
        -------
        array of shape (output_dim,) or dict of 1D arrays.
            An output vector. If the model has multiple outputs, a dictionary is returned instead.
        """
        # Auto-regressive mode
        if x is None:
            x = np.zeros((0,))
        check_model_timestep(x)

        if not self.initialized:
            self.initialize(x)

        x_mapping = map_input(self, x)

        state = {node: node.state for node in self.nodes}
        buffers = self.feedback_buffers
        new_buffers, new_state = self._step((buffers, state), x_mapping)

        for node in new_state:
            node.state = new_state[node]
        self.feedback_buffers = new_buffers

        if not self.is_multi_output:
            return new_state[self.outputs[0]]["out"]
        else:
            return {node.name: new_state[node]["out"] for node in self.outputs}

    def _run(
        self,
        state: tuple[FeedbackBuffers, Mapping[Node, State]],
        x: Mapping[Node, Timeseries],
    ) -> tuple[tuple[FeedbackBuffers, Mapping[Node, State]], Mapping[Node, Timeseries]]:

        buffers, node_states = state
        # can be run offline (can be faster) if there is no (buffer) feedback
        can_be_run_offline = len(buffers) == 0

        new_state: dict[Node, State] = {}

        if can_be_run_offline:
            output_timeseries: dict[Node, Timeseries] = {}
            # "offline" run: Node by Node
            for node in self.execution_order:
                inputs = []
                if node in x:
                    inputs.append(x[node])
                inputs += [output_timeseries[parent] for parent in self.parents[node]]
                node_input = np.concatenate(inputs, axis=-1)
                new_state[node], output_timeseries[node] = node._run(node_states[node], node_input)
        else:
            # "online" run: step by step
            n_timesteps = x[list(x.keys())[0]].shape[0]
            output_timeseries: dict[Node, Timeseries] = {
                node: np.zeros((n_timesteps, node.output_dim)) for node in self.nodes
            }
            for i, (timestep,) in enumerate(mapping_iterator(x)):
                buffers, node_states = self._step((buffers, node_states), timestep)

                for node in self.nodes:
                    output_timeseries[node][i] = node_states[node]["out"]

        return (buffers, new_state), output_timeseries

    def run(self, x: Optional[ModelInput] = None, iters: Optional[int] = None, workers: int = 1) -> ModelInput:
        """Run the Model on a sequence of data.

        Parameters
        ----------
        x : array ([s,] t, d), list of arrays (t, d) or a mapping of them, optional
            A timeseries, multiple timeseries, or a mapping of timeseries or multiple timeseries. Input of the Model.
        iters : int, optional
            If ``x`` is ``None``, a dimensionless timeseries of length ``iters``
            is used instead.
        workers : int, default to 1
            Number of workers used for parallelization. If set to -1, all available
            workers (threads or processes) are used.

        Returns
        -------
        array of shape ([n_inputs,] timesteps, output_dim) or list of arrays, or dict
            A sequence of output vectors. If the model has multiple outputs, a mapping is returned instead.
        """
        # Auto-regressive mode
        if x is None:
            x = np.zeros((iters, 0))
        check_model_input(x)

        if not self.initialized:
            self.initialize(x)

        x_mapping = map_input(self, x)

        previous_states = {node: node.state for node in self.nodes}
        previous_buffers = self.feedback_buffers
        if is_multiseries(x_mapping):
            result: dict[Node, list[Timeseries]] = defaultdict(list)
            iterable_x = unfold_mapping(x_mapping)

            p_output = Parallel(n_jobs=workers, require="sharedmem")(
                delayed(self._run)((previous_buffers, previous_states), timeseries) for timeseries in iterable_x
            )
            new_model_states, output = zip(*p_output)
            new_buffers, new_state = new_model_states[-1]
            for o in output:
                for node in o:
                    result[node].append(o[node])

        else:
            (new_buffers, new_state), result = self._run((previous_buffers, previous_states), x_mapping)

        for node in new_state:
            node.state = new_state[node]
        self.feedback_buffers = new_buffers

        if not self.is_multi_output:
            return result[self.outputs[0]]
        else:
            return {node.name: result[node] for node in self.outputs}

    def predict(self, x: Optional[ModelInput] = None, iters: Optional[int] = None, workers: int = 1) -> ModelInput:
        """Alias for :py:meth:`~.Model.run`.

        Parameters
        ----------
        x : array ([s,] t, d), list of arrays (t, d) or a mapping of them, optional
            A timeseries, multiple timeseries, or a mapping of timeseries or multiple timeseries. Input of the Model.
        iters : int, optional
            If ``x`` is ``None``, a dimensionless timeseries of length ``iters``
            is used instead.
        workers : int, default to 1
            Number of workers used for parallelization. If set to -1, all available
            workers (threads or processes) are used.

        Returns
        -------
        array of shape ([n_inputs,] timesteps, output_dim) or list of arrays, or dict
            A sequence of output vectors. If the model has multiple outputs, a mapping is returned instead.
        """
        return self.run(x=x, iters=iters, workers=workers)

    def _learning_step(
        self,
        state: tuple[FeedbackBuffers, dict[Node, State]],
        x: dict[Node, Timestep],
        y: dict[Node, Timestep],
    ) -> tuple[FeedbackBuffers, dict[Node, State]]:
        buffers, node_states = state
        new_state: dict[Node, State] = {}

        for node in self.execution_order:
            inputs = []
            if node in x:
                inputs.append(x[node])
            inputs += [buffer[-1] for (_p, _d, c), buffer in buffers.items() if c == node]
            inputs += [new_state[parent]["out"] for parent in self.parents[node]]
            node_input = np.concatenate(inputs, axis=-1)
            if isinstance(node, OnlineNode):
                node_target = y.get(node, None)
                new_state[node] = {"out": node._learning_step(node_input, node_target)}
            else:
                new_state[node] = node._step(node_states[node], node_input)

        new_buffers = {edge: buffer.copy() for edge, buffer in buffers.items()}
        for (p, d, c), buffer in new_buffers.items():
            buffer[-1] = new_state[p]["out"]
            new_buffers[(p, d, c)] = np.roll(buffer, 1, axis=0)

        return new_buffers, new_state

    def partial_fit(
        self,
        x: Union[Timeseries, dict[str, Timeseries]],
        y: Optional[Union[Timeseries, dict[str, Timeseries]]] = None,
    ) -> ModelInput:
        """Fit the Model in an online fashion.

        This method both trains the Model parameters and produce predictions on
        the run. Calling :py:meth:`partial_fit` updates the Model without
        resetting the parameters, unlike :py:meth:`fit`.

        Parameters
        ----------
        x : array-like of shape (timesteps, input_dim)
            Input sequence of data.
        y : array-like of shape (timesteps, output_dim), optional.
            Target sequence of data. If None, the Node will train in an
            unsupervised way, if possible.

        Returns
        -------
        array of shape (timesteps, output_dim) or mapping of arrays
            All outputs computed during the training.
        """
        if not self.is_online:
            raise TypeError("Trying to partial_fit a Model that can't be trained in an online manner.")

        check_model_input(x)
        if y is not None:
            check_model_input(y)

        if not self.initialized:
            self.initialize(x, y)

        x_ = map_input(self, x)
        y_ = map_teacher(self, y)

        n_timesteps = x_[list(x_.keys())[0]].shape[0]
        output_timeseries: dict[Node, Timeseries] = {
            node: np.zeros((n_timesteps, node.output_dim)) for node in self.nodes
        }

        states = {node: node.state for node in self.nodes}
        buffers = self.feedback_buffers
        for i, (x_timestep, y_timestep) in enumerate(mapping_iterator(x_, y_)):
            buffers, states = self._learning_step((buffers, states), x_timestep, y_timestep)
            for node in self.nodes:
                output_timeseries[node][i] = states[node]["out"]

        for node in states:
            node.state = states[node]
        self.feedback_buffers = buffers

        if not self.is_multi_output:
            return output_timeseries[self.outputs[0]]
        else:
            return {node.name: output_timeseries[node] for node in self.outputs}

    def fit(
        self,
        x: ModelInput,
        y: Optional[ModelInput] = None,
        warmup: int = 0,
        workers: int = 1,
    ) -> "Model":
        """Offline fitting method of a Model.

        Parameters
        ----------
        x : list or array-like of shape ([series, ] timesteps, input_dim) or a mapping of input, optional
            Input sequences dataset.
        y : list or array-like of shape ([series], timesteps, output_dim), or a mapping of input optional
            Teacher signals dataset. If None, the method will try to fit the
            Node in an unsupervised way, if possible.
        warmup : int, default to 0
            Number of timesteps to consider as warmup and
            discard at the beginning of each timeseries before training.

        Returns
        -------
        Node
            Node trained offline.
        """
        check_model_input(x)
        if y is not None:
            check_model_input(y)

        if not self.initialized:
            self.initialize(x, y)

        # TODO: try removing the default list
        result: dict[Node, NodeInput] = defaultdict(list)
        buffers = self.feedback_buffers

        x_ = map_input(self, x)
        y_ = map_teacher(self, y)

        # forced teaching
        for supervised in y_:
            # TODO: handle Unsupervised has children
            result[supervised] = y_[supervised]

        for node in self.pseudo_execution_order:
            inputs: list[NodeInput] = []
            if node in x_:
                inputs.append(x_[node])
            inputs += [result[parent] for parent in self.parents[node]]
            for (p, _d, c), buffer in buffers.items():
                if c == node:
                    new_buffer, data = data_from_buffer(buffer, result[p])
                    buffers[(p, _d, c)] = new_buffer
                    inputs.append(data)
            node_input = join_data(*inputs)
            if isinstance(node, TrainableNode):
                node_target = y_.get(node, None)
                if isinstance(node, ParallelNode):
                    node.fit(node_input, node_target, warmup=warmup, workers=workers)
                else:
                    node.fit(node_input, node_target, warmup=warmup)
            else:
                result[node] = node.run(node_input, workers=workers)

        return self

    def reset(self) -> tuple[dict[Node, State], FeedbackBuffers]:
        """Reset all Node states and buffers in the Model.

        Returns
        -------
        dict[str, np.array], dict[Edge, array]: previous states of the Nodes and previous feedback buffer values.
        """
        previous_node_states = {node: node.reset() for node in self.nodes}
        previous_buffers = self.feedback_buffers
        self.feedback_buffers = {k: np.zeros(v.shape) for k, v in self.feedback_buffers.items()}

        return previous_node_states, previous_buffers

    def __call__(self, x: Optional[ModelTimestep] = None) -> ModelTimestep:
        return self.step(x)

    def __str__(self):
        return f"Model({', '.join(map(str, self.nodes))})"

    def __repr__(self):
        return self.__str__()

    def __rshift__(self, other: Union[Node, "Model", Sequence[Union[Node, "Model"]]]) -> "Model":
        from .ops import link

        return link(self, other)

    def __rrshift__(self, other: Union[Node, "Model", Sequence[Union[Node, "Model"]]]) -> "Model":
        from .ops import link

        return link(other, self)

    def __lshift__(self, other: Union[Node, "Model", Sequence[Union[Node, "Model"]]]) -> "Model":
        from .ops import link_feedback

        return link_feedback(sender=other, receiver=self)

    def __rlshift__(self, other: Union[Node, "Model", Sequence[Union[Node, "Model"]]]) -> "Model":
        from .ops import link_feedback

        return link_feedback(sender=self, receiver=other)

    def __and__(self, other: Union[Node, "Model", Sequence[Union[Node, "Model"]]]) -> "Model":
        from .ops import merge

        return merge(self, other)

    def __rand__(self, other: Union[Node, "Model", Sequence[Union[Node, "Model"]]]) -> "Model":
        from .ops import merge

        return merge(other, self)



# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>

from __future__ import annotations

from typing import Sequence, TypeVar, Union

import jax
import numpy as np
from scipy.sparse import sparray

from reservoirpy.type import get_data_dimension, is_array, is_multiseries

global_dtype = jax.numpy.float64

# Creating a real type alias (Array1D) and then using it in another alias (Timestep)
# as a str is a trick to both benefits from type checks from pyright and the like
# and also have type aliases in documentation.
# See https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#confval-autodoc_type_aliases
# See also ../docs/source/conf.py:184
Array1D = Union[np.ndarray[tuple[int], np.dtype[np.floating]], jax.Array]
Array2D = Union[np.ndarray[tuple[int, int], np.dtype[np.floating]], jax.Array]
Array3D = Union[np.ndarray[tuple[int, int, int], np.dtype[np.floating]], jax.Array]

Timestep = Union["Array1D"]
Timeseries = Union["Array2D"]
MultiTimeseries = Union["Array3D", "Sequence[Timeseries]"]
NodeInput = Union[Timeseries, MultiTimeseries]
ModelInput = Union[NodeInput, dict[str, NodeInput]]
MappedTimestep = dict[str, Timestep]
ModelTimestep = Union[Timestep, MappedTimestep]
State = dict[str, jax.Array]
Edge = tuple["Node", int, "Node"]
Buffer = jax.Array

Weights = Union[np.ndarray, sparray, jax.Array]
Shape = TypeVar("Shape", int, tuple[int, ...])
FeedbackBuffers = dict[Edge, Buffer]



# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>

import jax.numpy as jnp
from numpy.testing import assert_array_equal, assert_equal

from reservoirpy.jax.nodes import Input, Output


def test_input():
    inp = Input()
    x = jnp.ones((10,))
    out = inp(x)
    assert_array_equal(out, x)
    x = jnp.ones((10, 10))
    out = inp.run(x)
    assert_array_equal(out, x)


def test_output():
    output = Output()
    x = jnp.ones((10,))
    out = output(x)
    assert_array_equal(out, x)
    x = jnp.ones((100, 10))
    out = output.run(x)
    assert_array_equal(out, x)



import jax.numpy as jnp
import scipy.sparse
from jax.experimental import sparse as jsparse

from .. import mat_gen


class JaxInitializer(mat_gen.Initializer):
    """Initializer for matrix generation for the Jax backend

    This inherits from the regular numpy/scipy Initializer class.
    The only difference is that the resulting array is converted to Jax
    (a regular :py:class:`jax.Array` or a :py:class:`jax.experimental.sparse.BCOO` array).
    """

    def __call__(self, *shape, **kwargs):
        call_result = mat_gen.Initializer.__call__(self, *shape, **kwargs)

        # Not initialized yet, only kwargs set
        if isinstance(call_result, mat_gen.Initializer):
            return call_result
        # Initialized, call_result is a sparse or dense array to convert to Jax
        else:
            if scipy.sparse.issparse(call_result):
                return jsparse.BCOO.from_scipy_sparse(call_result)
            else:
                return jnp.array(call_result)


bernoulli = JaxInitializer(mat_gen._bernoulli)
cluster = JaxInitializer(mat_gen._cluster)
line = JaxInitializer(mat_gen._line)
normal = JaxInitializer(mat_gen._normal)
orthogonal = JaxInitializer(mat_gen._orthogonal)
ring = JaxInitializer(mat_gen._ring)
small_world = JaxInitializer(mat_gen._small_world)
uniform = JaxInitializer(mat_gen._uniform)
zeros = JaxInitializer(mat_gen._zeros)
ones = JaxInitializer(mat_gen._ones)


def __getattr__(attr):
    if attr in ["random_sparse", "fast_spectral_initialization"]:
        raise AttributeError(f"{attr} does not have a Jax implementation. Please use reservoir.mat_gen.{attr} instead.")
    else:
        raise AttributeError(f"module {__name__!r} has no attribute {attr}")



"""
====================================================
ReservoirPy Jax Nodes (:mod:`reservoirpy.nodes.jax`)
====================================================

.. currentmodule:: reservoirpy.nodes.jax

Reservoirs
==========

.. autosummary::
    :toctree: generated/
    :template: autosummary/class.rst

    Reservoir - Recurrent pool of leaky integrator neurons
    NVAR - Non-linear Vector Autoregressive machine (NG-RC)
    IPReservoir - Reservoir with intrinsic plasticity learning rule
    LocalPlasticityReservoir - Reservoir with weight plasticity
    ES2N - Edge of Stability Echo State Network

Offline readouts
================

.. autosummary::
    :toctree: generated/
    :template: autosummary/class.rst

    Ridge - Layer of neurons connected through offline linear regression.

Online readouts
===============

.. autosummary::
    :toctree: generated/
    :template: autosummary/class.rst

    LMS - Layer of neurons connected through least mean squares learning rule.
    RLS - Layer of neurons connected through recursive least squares learning rule.

Activation functions
====================

.. autosummary::
    :toctree: generated/
    :template: autosummary/class.rst

    Tanh - Hyperbolic tangent node.
    Sigmoid - Logistic function node.
    Softmax - Softmax node.
    Softplus - Softplus node.
    ReLU - Rectified Linear Unit node.
    Identity - Identity function node.

Input and Output
================

.. autosummary::
    :toctree: generated/
    :template: autosummary/class.rst

    Input - Input node, used to distribute input data to other nodes.
    Output - Output node, used to gather stated from hidden nodes.

"""


# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>

from .activations import Identity, ReLU, Sigmoid, Softmax, Softplus, Tanh
from .es2n import ES2N
from .intrinsic_plasticity import IPReservoir
from .io import Input, Output
from .lif import LIF
from .lms import LMS

# from .local_plasticity_reservoir import LocalPlasticityReservoir
from .nvar import NVAR
from .reservoir import Reservoir
from .ridge import Ridge
from .rls import RLS

__all__ = [
    "Reservoir",
    "Input",
    "Output",
    "Ridge",
    "LMS",
    "RLS",
    "Tanh",
    "Softmax",
    "Softplus",
    "Identity",
    "Sigmoid",
    "ReLU",
    "NVAR",
    "IPReservoir",
    "LIF",
    # "LocalPlasticityReservoir",  # TODO: find a way to edit W as a sparse BCOO
    "ES2N",
]


def __getattr__(attr):
    if attr in ["LocalPlasticityReservoir", "ScikitLearnNode"]:
        raise AttributeError(f"{attr} does not have a Jax implementation. Please use reservoir.nodes.{attr} instead.")
    else:
        raise AttributeError(f"module {__name__!r} has no attribute {attr}")



# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>

from typing import Optional

from reservoirpy.type import Edge

from .model import Model
from .node import Node, TrainableNode
from .nodes import Input, Output, Reservoir, Ridge
from .utils.model_utils import obj_from_kwargs


class ESN(Model):
    """Simple Echo State Network.

    This class is provided as a wrapper for a simple reservoir connected to a
    readout.

    Parameters
    ----------
    units : int, optional
        Number of reservoir units. If None, the number of units will be inferred from
        the ``W`` matrix shape.
    reservoir : Node, optional
        A Node instance to use as a reservoir,
        such as a :py:class:`~reservoirpy.nodes.Reservoir` node.
    readout : Node, optional
        A Node instance to use as a readout,
        such as a :py:class:`~reservoirpy.nodes.Ridge` node
        (only this one is supported).
    feedback : bool, defaults to False
        If True, the readout is connected to the reservoir through
        a feedback connection.
    input_to_readout : bool, defaults to False
        If True, the input is directly fed to the readout. See
        :ref:`/user_guide/advanced_demo.ipynb#Input-to-readout-connections`.
    return_reservoir_activity : bool, defaults to False
        If True, the model outputs a dict with the reservoir activity in addition to the readout output.
    **kwargs
        Arguments passed to the reservoir and readout.

    See Also
    --------
    Reservoir
    Ridge

    Example
    -------
    >>> from reservoirpy import ESN
    >>>
    >>> model = ESN(units=100, sr=0.9, ridge=1e-6)  # reservoir and readout parameters at once
    >>> model.fit(x_train, y_train)
    >>>
    >>> model = ESN(reservoir=Reservoir(100, sr=0.9), readout=Ridge(1e-5))  # passing nodes as parameters
    >>>
    >>> model = ESN(units=100, input_to_readout=True, feedback=True)  # more complex model
    """

    #: A :py:class:`~reservoirpy.nodes.Reservoir` or a :py:class:`~reservoirpy.nodes.NVAR` instance.
    reservoir: Node
    #: A :py:class:`~reservoirpy.nodes.Ridge` instance.
    readout: TrainableNode
    #: Is readout connected to reservoir through feedback (False by default).
    feedback: bool
    #: Is the readout directly receiving the input (False by default).
    input_to_readout: bool
    #: Are the reservoir states returned by the model along the readout output (False by default).
    return_reservoir_activity: bool
    #: Output node for the reservoir, if ``return_reservoir_activity`` is set to True
    output_node: Optional[Output] = None
    #: Input node, if ``input_to_readout`` is set to True
    input_node: Optional[Input] = None

    def __init__(
        self,
        reservoir: Optional[Reservoir] = None,
        readout: Optional[Ridge] = None,
        feedback: bool = False,
        input_to_readout: bool = False,
        return_reservoir_activity: bool = False,
        **kwargs,
    ):
        if "name" in kwargs:
            kwargs.pop("name")

        if reservoir is None:
            reservoir = obj_from_kwargs(Reservoir, kwargs)

        if readout is None:
            # avoid argument name collision
            if "bias" in kwargs:
                kwargs.pop("bias")
            if "readout_bias" in kwargs:
                kwargs["bias"] = kwargs["readout_bias"]
                kwargs.pop("readout_bias")
            if "input_dim" in kwargs:
                kwargs.pop("input_dim")
            kwargs["name"] = "readout"
            readout = obj_from_kwargs(Ridge, kwargs)

        nodes: list[Node] = [reservoir, readout]
        edges: list[Edge] = [(reservoir, 0, readout)]

        if feedback:
            edges.append((readout, 1, reservoir))

        if input_to_readout:
            input_node = Input()
            nodes.append(input_node)
            edges.append((input_node, 0, reservoir))
            edges.append((input_node, 0, readout))
            self.input_node = input_node

        if return_reservoir_activity:
            output_node = Output(name="reservoir")
            nodes.append(output_node)
            edges.append((reservoir, 0, output_node))
            self.output_node = output_node

        self.reservoir = reservoir
        self.readout = readout
        self.feedback = feedback
        self.input_to_readout = input_to_readout
        self.return_reservoir_activity = return_reservoir_activity

        super(ESN, self).__init__(nodes=nodes, edges=edges)



# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>

import jax
import jax.numpy as jnp
import pytest
from numpy.testing import assert_array_almost_equal, assert_array_equal

from reservoirpy.jax.activationsfunc import relu
from reservoirpy.jax.mat_gen import bernoulli, zeros
from reservoirpy.jax.nodes import ES2N


def test_es2n_init():
    node = ES2N(100, proximity=0.8, bias=zeros)

    data = jnp.ones((10,))
    res = node(data)

    assert node.W.shape == (100, 100)
    assert node.Win.shape == (100, 10)
    assert node.proximity == 0.8
    assert node.units == 100

    data = jnp.ones((120, 10))
    res = node.run(data)

    assert res.shape == (120, 100)

    with pytest.raises(ValueError):
        ES2N()

    res = ES2N(100, activation="relu")
    assert id(res.activation) == id(jax.nn.relu)


def test_es2n_init_from_proximity_is_arrays():
    proximity = jnp.ones((100,)) * 0.5
    input_scaling = jnp.ones((10,)) * 0.8
    node = ES2N(100, proximity=proximity, input_scaling=input_scaling)

    data = jnp.ones((2, 10))
    res = node.run(data)

    assert node.W.shape == (100, 100)
    assert node.Win.shape == (100, 10)
    assert_array_equal(node.proximity, jnp.ones(100) * 0.5)
    assert_array_equal(node.input_scaling, jnp.ones(10) * 0.8)


def test_es2n_init_from_matrices():
    Win = jnp.ones((100, 10))

    node = ES2N(100, proximity=0.8, Win=Win, bias=bernoulli)

    data = jnp.ones((10,))
    res = node(data)

    assert node.W.shape == (100, 100)
    assert_array_equal(node.Win, Win)
    assert node.proximity == 0.8
    assert node.units == 100

    data = jnp.ones((10000, 10))
    res = node.run(data)

    assert res.shape == (10000, 100)

    Win = jnp.ones((100, 10))
    bias = jnp.ones((100,))

    node = ES2N(100, proximity=0.8, Win=Win, bias=bias)

    data = jnp.ones((10,))
    res = node(data)

    assert node.W.shape == (100, 100)
    assert node.proximity == 0.8
    assert node.units == 100
    assert node.output_dim == 100

    data = jnp.ones((10000, 10))
    res = node.run(data)

    assert res.shape == (10000, 100)

    with pytest.raises(ValueError):  # Shape override (matrix.shape > units parameter)
        data = jnp.ones((10))
        W = jnp.ones((10, 10))
        res = ES2N(100, W=W)
        _ = res(data)

    with pytest.raises(TypeError):  # Bad matrix shape
        W = jnp.ones((10, 11))
        res = ES2N(W=W)
        res(data)

    with pytest.raises(ValueError):  # Bad matrix format
        res = ES2N(100, W=1.0)
        res(data)

    with pytest.raises((ValueError, TypeError)):  # Bad Win shape
        res = ES2N(100, Win=jnp.ones((100, 20)))
        res(data)

    with pytest.raises((ValueError, TypeError)):  # Bad Win shape
        res = ES2N(100, Win=jnp.ones((101, 10)))
        res(data)

    with pytest.raises(ValueError):  # Bad matrix format
        res = ES2N(100, Win=1.0)
        res(data)


def test_es2n_bias():
    node = ES2N(100, proximity=0.8, bias=zeros)
    data = jnp.ones((10,))
    res = node(data)
    assert node.W.shape == (100, 100)
    assert node.Win.shape == (100, 10)
    assert node.bias.shape == (100,)
    assert_array_equal(node.bias, jnp.zeros((100,)))
    assert node.proximity == 0.8
    assert node.units == 100

    node = ES2N(100, proximity=0.8, bias=bernoulli)
    data = jnp.ones((10,))
    res = node(data)
    assert node.bias.shape == (100,)

    bias = jnp.ones((100,))
    node = ES2N(100, bias=bias)
    res = node(data)
    assert_array_equal(node.bias, bias)

    bias = jnp.ones((100,))
    node = ES2N(100, bias=bias)
    res = node(data)
    assert_array_equal(node.bias, bias)

    data = jnp.zeros((100,))
    node = ES2N(100, bias=1.0)
    res = node(data)
    assert_array_equal(res, jnp.tanh(jnp.ones((100,))))

    with pytest.raises(AssertionError):
        bias = jnp.ones((100, 1))
        node = ES2N(100, bias=bias)
        res = node(data)
        assert res.shape == (100,)

    with pytest.raises(ValueError):
        bias = jnp.ones((100, 2))
        node = ES2N(100, bias=bias)
        node(data)


def test_es2n_run():
    x = jnp.ones((10, 5))

    res = ES2N(100)
    out = res.run(x)
    assert out.shape == (10, 100)


def test_es2n_chain():
    node1 = ES2N(100, proximity=0.8)
    node2 = ES2N(50, proximity=1.0)

    data = jnp.ones((10,))
    res = (node1 >> node2)(data)

    assert node1.W.shape == (100, 100)
    assert node1.Win.shape == (100, 10)
    assert node2.W.shape == (50, 50)
    assert node2.Win.shape == (50, 100)

    assert res.shape == (50,)


def test_es2n_seed():
    node1 = ES2N(100, seed=123)
    node2 = ES2N(100, seed=123)

    data = jnp.ones((10, 10))

    assert_array_almost_equal(node1.run(data), node2.run(data))



# # Licence: MIT License
# # Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>

# from typing import Callable, Literal, Optional, Sequence, Union

# from jax.experimental.sparse import BCOO
# import jax.numpy as jnp
# from numpy.random import Generator

# from reservoirpy.utils.data_validation import check_node_input

# from ..activationsfunc import get_function, tanh
# from ..mat_gen import bernoulli, uniform
# from ..node import TrainableNode
# from ..type import (
#     NodeInput,
#     State,
#     Timeseries,
#     Timestep,
#     Weights,
# )
# from ...type import is_array, is_multiseries
# from ..utils import rand_generator


# class LocalPlasticityReservoir(TrainableNode):
#     """
#     A reservoir that learns its recurrent weights W through a local
#     learning rule selected by the ``learning_rule`` hyperparameter.

#     Reservoir states are updated with the external equation:

#     .. math::

#         & r[t+1] = (1 - lr)*r[t] + lr*(W r[t] + W_{in} u[t+1] + bias) \\\\
#         & x[t+1] = f(r[t+1])

#     Where :math:`f` is the activation function.

#     Then the local rule is applied each timestep to update W.

#     .. math::

#         W_{ij} \\leftarrow W_{ij} + \\Delta W_{ij}

#     Supported rules:
#       ``oja``:
#         :math:`\\Delta W_{ij} = \\eta y (x - y W_{ij})`
#       ``anti-oja`` [1]_ [2]_ [3]_ :
#         :math:`\\Delta W_{ij} = - \\eta y (x - y W_{ij})`
#       ``hebbian`` [4]_ :
#         :math:`\\Delta W_{ij} = \\eta x y`
#       ``anti-hebbian``:
#         :math:`\\Delta W_{ij} = - \\eta x y`
#       ``bcm`` [2]_ :
#         :math:`\\Delta W_{ij} = \\eta x y (y - \\theta_{BCM})`

#     Where :math:`x` represents the pre-synaptic state and :math:`y` represents
#     the post-synaptic state of the neuron.

#     For "`bcm`", you can set a threshold ``bcm_theta`` (default `0.0`).

#     If ``synapse_normalization=True``, then after each local-rule update
#     on a row i of W, the row is rescaled to unit L2 norm. [4]_


#     Parameters
#     ----------
#     units : int, optional
#         Number of reservoir units. If None, the number of units will be inferred from
#         the ``W`` matrix shape.
#     local_rule : str, default to `oja`
#         One of `"oja"`, `"anti-oja"`, `"hebbian"`, `"anti-hebbian"`, `"bcm"`.
#     eta : float, default to 1e-3.
#         Local learning rate for the weight update.
#     bcm_theta : float, default to 0.0
#         The threshold used in the "bcm" rule.
#     synapse_normalization : bool, default to True
#         If True, L2-normalize each row of W after its update.
#     epochs : int, default to 1
#         Number of training iterations.
#     sr : float, default to 1.0
#         Spectral radius of recurrent weight matrix.
#     lr : float or array-like of shape (units,), default to 1.0
#         Neurons leak rate. Must be in :math:`[0, 1]`.
#     input_scaling : float or array-like of shape (features,), default to 1.0.
#         Input gain. An array of the same dimension as the inputs can be used to
#         set up different input scaling for each feature.
#     input_connectivity : float, default to 0.1
#         Connectivity of input neurons, i.e. ratio of input neurons connected
#         to reservoir neurons. Must be in :math:`]0, 1]`.
#     rc_connectivity : float, default to 0.1
#         Connectivity of recurrent weight matrix, i.e. ratio of reservoir
#         neurons connected to other reservoir neurons, including themselves.
#         Must be in :math:`]0, 1]`.
#     Win : callable or array-like of shape (units, features), default to :py:func:`~reservoirpy.mat_gen.bernoulli`
#         Input weights matrix or initializer. If a callable (like a function) is
#         used,
#         then this function should accept any keywords
#         parameters and at least two parameters that will be used to define the
#         shape of
#         the returned weight matrix.
#     W : callable or array-like of shape (units, units), default to :py:func:`~reservoirpy.mat_gen.uniform`
#         Recurrent weights matrix or initializer. If a callable (like a function) is
#         used, then this function should accept any keywords
#         parameters and at least two parameters that will be used to define the
#         shape of
#         the returned weight matrix.
#     bias : callable or array-like of shape (units, 1), default to :py:func:`~reservoirpy.mat_gen.bernoulli`
#         Bias weights vector or initializer. If a callable (like a function) is
#         used, then this function should accept any keywords
#         parameters and at least two parameters that will be used to define the
#         shape of
#         the returned weight matrix.
#     activation : {"tanh", "sigmoid"}, default to "tanh"
#         Reservoir units activation function.
#     input_dim : int, optional
#         Input dimension. Can be inferred at first call.
#     seed : int or :py:class:`numpy.random.Generator`, optional
#         A random state seed, for noise generation.
#     dtype : Numpy dtype, default to jnp.float64
#         Numerical type for node parameters.
#     name : str, optional
#         Node name.


#     References
#     ----------

#     .. [1] Babinec, Š., & Pospíchal, J. (2007). Improving the prediction
#            accuracy of echo state neural networks by anti-Oja’s learning.
#            In International Conference on Artificial Neural Networks (pp. 19-28).
#            Berlin, Heidelberg: Springer Berlin Heidelberg.
#            https://doi.org/10.1007/978-3-540-74690-4_3

#     .. [2] Yusoff, M. H., Chrol-Cannon, J., & Jin, Y. (2016). Modeling neural
#            plasticity in echo state networks for classification and regression.
#            Information Sciences, 364, 184-196.
#            https://doi.org/10.1016/j.ins.2015.11.017

#     .. [3] Morales, G. B., Mirasso, C. R., & Soriano, M. C. (2021). Unveiling
#            the role of plasticity rules in reservoir computing. Neurocomputing,
#            461, 705-715. https://doi.org/10.1016/j.neucom.2020.05.127

#     .. [4] Wang, X., Jin, Y., & Hao, K. (2021). Synergies between synaptic and
#            intrinsic plasticity in echo state networks. Neurocomputing,
#            432, 32-43. https://doi.org/10.1016/j.neucom.2020.12.007

#     Example
#     -------
#     >>> reservoir = LocalPlasticityReservoir(
#     ...     units=100, sr=0.9, local_rule="hebbian",
#     ...     eta=1e-3, epochs=5, synapse_normalization=True
#     ... )
#     >>> # Fit on data timeseries
#     >>> reservoir.fit(X_data, warmup=10)
#     >>> # Then run
#     >>> states = reservoir.run(X_data)
#     """

#     #: Number of neuronal units in the reservoir.
#     units: int
#     #: Local learning rate for the weight update.
#     eta: float
#     #: The threshold used in the "bcm" rule.
#     bcm_theta: float
#     #: If True, L2-normalize each row of W after its update.
#     synapse_normalization: bool
#     #: Number of training iterations.
#     epochs: int
#     #: Leaking rate (1.0 by default) (:math:`\mathrm{lr}`).
#     lr: float
#     #: Spectral radius of ``W`` (optional).
#     sr: float
#     #: Input scaling (float or array) (1.0 by default).
#     input_scaling: Union[float, Sequence]
#     #: Connectivity (or density) of ``Win`` (0.1 by default).
#     input_connectivity: float
#     #: Connectivity (or density) of ``Wfb`` (0.1 by default).
#     rc_connectivity: float
#     #: Input weights matrix (:math:`\mathbf{W}_{in}`).
#     Win: Weights
#     #: Recurrent weights matrix (:math:`\mathbf{W}`). Must be sparse.
#     W: Weights
#     #: Bias vector (:math:`\mathbf{b}`).
#     bias: Weights
#     #: Activation of the reservoir units (tanh by default) (:math:`f`).
#     activation: Callable
#     #: Type of matrices elements. By default, ``jnp.float64``.
#     dtype: type
#     #: A random state generator. Used for generating Win and W.
#     rng: Generator

#     def __init__(
#         self,
#         units: Optional[int] = None,
#         # local rule choice
#         local_rule: Literal["oja", "anti-oja", "hebbian", "anti-hebbian", "bcm"] = "oja",
#         eta: float = 1e-3,
#         bcm_theta: float = 0.0,
#         synapse_normalization: bool = False,
#         epochs: int = 1,
#         # standard reservoir params
#         sr: float = 1.0,
#         lr: float = 1.0,
#         input_scaling: Union[float, Sequence] = 1.0,
#         input_connectivity: float = 0.1,
#         rc_connectivity: float = 0.1,
#         Win: Union[Weights, Callable] = bernoulli,
#         W: Union[Weights, Callable] = uniform,
#         bias: Union[Weights, Callable] = bernoulli,
#         activation: Union[str, Callable] = tanh,
#         input_dim: Optional[int] = None,
#         seed: Optional[Union[int, Generator]] = None,
#         dtype: type = jnp.float64,
#         name: Optional[str] = None,
#     ):
#         self.units = units
#         self.eta = eta
#         self.bcm_theta = bcm_theta
#         self.synapse_normalization = synapse_normalization
#         self.epochs = epochs
#         self.lr = lr
#         self.sr = sr
#         self.input_scaling = input_scaling
#         self.input_connectivity = input_connectivity
#         self.rc_connectivity = rc_connectivity
#         self.Win = Win
#         self.W = W
#         self.bias = bias
#         self.activation = get_function(activation)
#         self.dtype = dtype
#         self.rng = rand_generator(seed=seed)
#         self.name = name

#         # set units / output_dim
#         if units is None and not is_array(W):
#             raise ValueError("'units' parameter must not be None if 'W' parameter is not a matrix.")
#         if units is not None and is_array(W) and W.shape[-1] != units:
#             raise ValueError(
#                 f"Both 'units' and 'W' are set but their dimensions doesn't match: " f"{units} != {W.shape[-1]}."
#             )
#         self.units = units if units is not None else W.shape[-1]
#         self.output_dim = self.units

#         # set input_dim (if possible)
#         if input_dim is not None and is_array(Win) and Win.shape[-1] != input_dim:
#             raise ValueError(
#                 f"Both 'input_dim' and 'Win' are set but their dimensions doesn't "
#                 f"match: {input_dim} != {Win.shape[-1]}."
#             )
#         self.input_dim = Win.shape[-1] if is_array(Win) else input_dim

#         # Validate local rule name
#         local_rule = local_rule.lower()

#         def oja(weights, pre_state, post_state):
#             return eta * post_state * (pre_state - post_state * weights)

#         def anti_oja(weights, pre_state, post_state):
#             return -eta * post_state * (pre_state - post_state * weights)

#         def hebbian(weights, pre_state, post_state):
#             return eta * post_state * pre_state

#         def anti_hebbian(weights, pre_state, post_state):
#             return -eta * post_state * pre_state

#         def bcm(weights, pre_state, post_state):
#             return eta * post_state * (post_state - bcm_theta) * pre_state

#         rules = {
#             "oja": oja,
#             "anti-oja": anti_oja,
#             "hebbian": hebbian,
#             "anti-hebbian": anti_hebbian,
#             "bcm": bcm,
#         }
#         if not local_rule in rules:
#             raise ValueError(
#                 f"Unknown learning rule '{local_rule}'. Choose from: "
#                 "['oja', 'anti-oja', 'hebbian', 'anti-hebbian', 'bcm']."
#             )

#         self.increment = rules[local_rule]

#     def initialize(self, x: Optional[Union[NodeInput, Timestep]]):

#         # set input_dim
#         self._set_input_dim(x)

#         [Win_rng, W_rng, bias_rng] = self.rng.spawn(3)

#         if callable(self.Win):
#             self.Win = self.Win(
#                 self.units,
#                 self.input_dim,
#                 input_scaling=self.input_scaling,
#                 connectivity=self.input_connectivity,
#                 dtype=self.dtype,
#                 seed=Win_rng,
#             )

#         if callable(self.W):
#             self.W = self.W(
#                 self.units,
#                 self.units,
#                 sr=self.sr,
#                 connectivity=self.rc_connectivity,
#                 dtype=self.dtype,
#                 seed=W_rng,
#                 sparsity_type="coo",
#             )

#         if callable(self.bias):
#             self.bias = self.bias(
#                 self.units,
#                 connectivity=1.0,
#                 dtype=self.dtype,
#                 seed=bias_rng,
#             )

#         self.state = {
#             "out": jnp.zeros((self.units,)),
#             "internal": jnp.zeros((self.units,)),
#         }

#         if not isinstance(self.W, BCOO):
#             raise TypeError(f"Weight matrix W should be a BCOO matrix, got {type(self.W)}")
#         self.state = {"out": jnp.zeros((self.output_dim,))}

#         self.initialized = True

#     def _step(self, state: State, x: Timestep) -> State:
#         W = self.W  # NxN
#         Win = self.Win  # NxI
#         bias = self.bias  # N or float
#         f = self.activation
#         lr = self.lr
#         s = state["out"]

#         next_state = W @ s + Win @ x + bias
#         next_state = (1 - lr) * s + lr * next_state

#         return {"internal": next_state, "out": f(next_state)}

#     def fit(self, x: NodeInput, y: None = None, warmup: int = 0) -> "LocalPlasticityReservoir":
#         check_node_input(x, expected_dim=self.input_dim)

#         if not self.initialized:
#             self.initialize(x)

#         increment = self.increment
#         do_norm = self.synapse_normalization

#         def _local_synaptic_plasticity(seq: Timeseries):
#             """
#             Apply the local learning rule (Oja, Anti-Oja, Hebbian, Anti-Hebbian, BCM)
#             to update the recurrent weight matrix W.

#             If `synapse_normalization=True`, then each row of W is L2-normalized
#             immediately after the local rule update.
#             """
#             for u in seq:
#                 pre_state = self.state["internal"]  # (units,)
#                 post_state = self._step(self.state, u)["internal"]  # (units,)
#                 # Vectorized update of nonzero elements based on the chosen rule.
#                 (rows, cols, data) = self.W.indices[:, 0], self.W.indices[:, 1], self.W.data
#                 self.W.at[rows, cols].add(increment(data, pre_state[cols], post_state[rows]))
#                 # Optionally normalize each row.
#                 if do_norm:
#                     # Compute the L2 norm per row for the updated data.
#                     row_norms = jnp.sqrt(jnp.sum(self.W**2, axis=1)).reshape(-1, 1)
#                     safe_norms = jnp.where(row_norms > 0, row_norms, 1)
#                     self.W /= safe_norms

#         for _epoch in range(self.epochs):
#             if is_multiseries(x):
#                 for seq in x:
#                     _local_synaptic_plasticity(seq[warmup:])
#             else:
#                 _local_synaptic_plasticity(x[warmup:])
#         return self
